""" bob.py - Big-O-Bundler, hoff.industrie's release and build system
    Copyright (C) 2024  hoff.industries

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>."""

import sys
import subprocess

from os import path, mkdir, walk, makedirs, getcwd
from shutil import copyfile
from pathlib import Path

from argparse import ArgumentParser
from enum import Enum
from json import loads, decoder

BOB_URL = "https://hoff.industries/release-bob.html"
BOB_TITLE = "Big-O-Bundler"

COLOURS = {
	"clear": "\033[0m",
	"minion-yellow": "\033[38;5;220m",
	"ruby-red": "\033[38;5;196m",
	"spring-green": "\033[38;5;47m",
	"orange-orange": "\033[38;5;202m"
}

FLAVOUR_TEXT = f"{COLOURS['minion-yellow']}bob.py>{COLOURS['clear']}"
ERROR_TEXT = f"{COLOURS['ruby-red']}bob.py>{COLOURS['clear']}"
SUCCESS_TEXT = f"{COLOURS['spring-green']}bob.py>{COLOURS['clear']}"
WARNING_TEXT = f"{COLOURS['orange-orange']}bob.py>{COLOURS['clear']}"

class ExitCode(Enum):
	BAD_BOB_JSON = 1,
	BAD_JSON_DECODE = 2,
	BAD_TARGET = 3

RESERVED = [
	"*all*",
	"*name*",
	"*exclude*"
]

def main():

	print(f"{FLAVOUR_TEXT} {BOB_TITLE} (https://banjokazooiewiki.com/wiki/Big-O-Blaster)")

	targets = {}
	target_listings = {}

	parser = ArgumentParser(
		prog="bob",
		description="hoff.industrie's release and build system"
	)

	parser.add_argument("-d", "--dirname",
		help="directory to look for targets in", default="./")

	parser.add_argument("-t", "--target",
		help="release/build target", default="*all*")

	parser.add_argument("-g", "--ball",
		help="tarball release after copy", action="store_true")

	parser.add_argument("-v", "--version",
		help="custom version number for release/build", default="*hash*")

	parser.add_argument("-r", "--release",
		action="store_true", help="package target for release")

	parser.add_argument("-b", "--build",
		action="store_true", help="try build target")

	parser.add_argument("-p", "--project",
		action="store_true", help="build entire project")

	arguments = parser.parse_args()

	if arguments.dirname == "./":
		my_path = getcwd()
	else:
		my_path = path.abspath(arguments.dirname)

	print(f"{FLAVOUR_TEXT} working in {my_path}")

	try:

		with open(path.join(arguments.dirname, "bob.json"), "r") as bob_file:
			targets = loads(bob_file.read())
	except FileNotFoundError:

		print(f"{ERROR_TEXT} error: couldn't find bob.json metadata file!", file=sys.stderr)
		sys.exit(ExitCode.BAD_BOB_JSON.value)
	except decoder.JSONDecodeError as e:

		print(f"{ERROR_TEXT} error: Couldn't decode <bob.json>! {e}")
		sys.exit(ExitCode.BAD_JSON_DECODE.value)

	if arguments.version == "*hash*":

		git_ps = subprocess.run(["git", "-C", arguments.dirname,
			"rev-parse", "--short", "HEAD"], capture_output=True)
		version = git_ps.stdout.decode("ASCII").strip()

		print(f"{FLAVOUR_TEXT} Found git hash <{version}>.")
	else:
		version = arguments.version

	git_ps = subprocess.run(["git", "-C", arguments.dirname,
		"branch", "--show-current"], capture_output=True)
	my_branch = git_ps.stdout.decode("ASCII").strip()

	print(f"{FLAVOUR_TEXT} Found git branch <{my_branch}>.")

	git_ps = subprocess.run(["git", "-C", arguments.dirname,
		"status", "--porcelain"], capture_output=True)

	if len(git_ps.stdout) != 0:
		print(f"{WARNING_TEXT} tree is MODIFIED I'll add the '-w' tag.")
		version += "-w"

	print(f"{FLAVOUR_TEXT} loaded <bob.json> on <{my_branch}> version <{version}>")
	print(f"{FLAVOUR_TEXT} targeting <{arguments.target}>")

	releases_path = path.join(arguments.dirname, "release")

	try:
		mkdir(releases_path)
		print(f"{FLAVOUR_TEXT} made releases path {releases_path}")
	except FileExistsError:
		print(f"{WARNING_TEXT} using existing releases path {releases_path}")

	if arguments.target == "*all*":
		for target in targets:

			if target in RESERVED:
				continue

			target_listings[target] = targets[target] + targets["*all*"]
			print(f"{FLAVOUR_TEXT} added target <{target}>")

	else:
		try:
			target_listings[arguments.target] = targets[arguments.target] + targets["*all*"]
			print(f"{FLAVOUR_TEXT} added target <{arguments.target}>")
		except KeyError:
			print(f"{ERROR_TEXT} error: no target found <{arguments.target}>")
			sys.exit(ExitCode.BAD_TARGET.value)

	for target in target_listings:
		include_list = target_listings[target]
		target_release_path = path.join(releases_path, f"{target}-{my_branch}-{version}")

		try:
			mkdir(target_release_path)
			print(f"{FLAVOUR_TEXT} made path {target_release_path}")
		except FileExistsError:
			print(f"{WARNING_TEXT} using existing path {target_release_path}")

		for include in include_list:
			source_path = path.join(arguments.dirname, include)
			destination_path = path.join(target_release_path, include)

			if path.isfile(source_path):
				print(f"{FLAVOUR_TEXT} copying regular file {include}")
				copyfile(source_path, destination_path)
				continue

			for root, dirs, files in walk(source_path):
				for include_file in files:

					if include_file[0] == ".":
						print(f"{WARNING_TEXT} skipping hidden file {include_file}")
						continue

					target_path = root.replace(f"{my_path}", "")
					if target_path[0] == "/":
						target_path = target_path[1:]

					destination_path = path.join(target_release_path, target_path)

					try:
						makedirs(destination_path)
						print(f"{FLAVOUR_TEXT} made path {destination_path}")
					except FileExistsError:
						pass

					if Path(destination_path).name == "build":
						print(f"{WARNING_TEXT} skipping build file {include_file}")
						continue

					copyfile(path.join(root, include_file),
						path.join(destination_path, include_file))

					print(include_file)

		with open(path.join(target_release_path, ".bundled-by-bob"), "w") as calling_card:
			calling_card.write(f"this release <{target}-{my_branch}-{version}> was generated by B.O.B. <{BOB_TITLE}> {BOB_URL}")

		print(f"{SUCCESS_TEXT} release <{version}> for <{target}> done to {target_release_path}")

		if arguments.ball:
			subprocess.run(["tar", "-czf", f"{target_release_path}.tar.gz", "-C", target_release_path, "."])
			print(f"{SUCCESS_TEXT} tarball'ed <{version}> for <{target}> to {target_release_path}.tar.gz")

if __name__ == "__main__":
	main()
